// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: topology.cc,v 1.5 2003/03/03 21:57:00 jewillco Exp $
//
// The Cartesian function creates an intra communicator with a Cartesian
// topology.  Then several of the OOMPI_Cart_comm member functions are
// called to demonstrate that it is working correctly.  The graph
// function creates an intra comm with a graph topology.  The graph it
// creates has 4 nodes, so you will need to invoke the program with at
// least 4 processes.  Then to demonstrate that it is working correctly,
// a couple member functions of OOMPI_Graph_comm are called.


#include <iostream> 
#include "oompi.h" 

using namespace std;

void cartesian();
void graph();

int my_rank;


int
main(int argc, char *argv[])
{
  OOMPI_COMM_WORLD.Init(argc, argv);
  my_rank = OOMPI_COMM_WORLD.Rank();

  if (OOMPI_COMM_WORLD.Size() >= 4) {
    cartesian();
    graph();
  }
  else 
    if (my_rank == 0)
      cout << "This example must be run with at least 4 nodes." << endl;

  OOMPI_COMM_WORLD.Finalize();
}
  
void 
cartesian()
{
  if (my_rank == 0)
    cout << endl << "CARTESIAN TOPOLOGY" << endl << endl;

  //  Create a 2-D mesh which is periodic in both dimensions.  Dims_create()
  //  will be used to set the number of nodes per dimension.  Periodic is
  //  set true, and Reorder is set false.  Note: if we wanted one
  //  dimension to be periodic and the other not, then an array of
  //  booleans would have to be created and passed in.  When all the
  //  dimensions have the same periodicity, then a single boolean can be
  //  used.

  int dims[2] = {0,0};

  OOMPI_Cart_comm cart(OOMPI_COMM_WORLD, 2, 
		       OOMPI_COMM_WORLD.Dims_create(2,dims), true, false);

  cout << "cart is a " << dims[0] << "x" << dims[1] << " 2D mesh" << endl;

  // Find my left and right neighbors
  int right, left;
  right = cart.Shift(1, 1, left);
  
  // Find my coordinates in the mesh
  int *coords = cart.Coords();
  cout << "Process " << my_rank << " is at coords: " << coords[0] << " "
    << coords[1] << "." << endl << "To the left is node " 
      << left << ". to the right is node " << right << "." << endl;

  // Free coords
  delete [] coords;
}

void
graph()
{
  if (my_rank == 0)
    cout << endl << "GRAPH TOPOLOGY" << endl << endl;

  // Create a graph with these connections 

  // process   | neighbors
  // ---------------------
  //    0      | 1,3
  //    1      | 0
  //    2      | 3
  //    3      | 0,2

  // Make a communicator with only the first four ranks in it

  OOMPI_Intra_comm newcomm = OOMPI_COMM_WORLD.Split((my_rank < 4), my_rank);

  if (my_rank < 4) {
    
    int edges[] = {1,3,0,3,0,2};
    int index[] = {2,3,4,6};
    
    OOMPI_Graph_comm graph(newcomm, 4, index, edges, false);
    
    for (int i = 0; i < graph.Size(); i++) {
      int *neighbors = graph.Neighbors(i);
      cout << "Node " << i << " is connected to : ";
      for (int j = 0; j < graph.Neighbors_count(i); j++)
	cout << neighbors[j] << " ";
      cout << endl;
      delete [] neighbors;
    }
  }

  OOMPI_COMM_WORLD.Barrier();
}
