// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: graph_test.cc,v 1.4 2003/02/13 21:07:48 jsquyres Exp $
//
// OOMPI test program
// Test OOMPI_Graph functionality
//

#include "oompi.h"
#include "oompi_test.h"


void 
OOMPI_Graph_comm_test(void)
{
  Testing("OOMPI_Graph_comm");

  MPI_Comm mpi_graph_comm;

  {
  OOMPI_Graph_comm D;
  int x_nedges = comm_size * 2;
  int x_nnodes = comm_size;
  int *x_index = new int [x_nnodes];
  int *x_edges = new int [x_nedges];

  Testing("OOMPI_Graph_comm::Constructors");
  {
    Testing("OOMPI_Graph_comm::normal constructor");
    for (int i=0; i<x_nnodes; i++) {
      x_index[i] = (i+1)*2;
      x_edges[i*2] = (i - 1 + comm_size) % comm_size;
      x_edges[i*2+1] = (i + 1) % comm_size;
    }
    
    MPI_Graph_create(MPI_COMM_WORLD, x_nnodes, x_index, 
		     x_edges, 0, &mpi_graph_comm);

    OOMPI_Graph_comm A(OOMPI_COMM_WORLD, x_nnodes, x_index, x_edges);
    int compare;

    MPI_Comm_compare(A.Get_mpi(), mpi_graph_comm, &compare);
    Test(compare == MPI_CONGRUENT);
    
    
    Testing("OOMPI_Graph_comm::mpi constructor");
    OOMPI_Graph_comm B(mpi_graph_comm);
    Test(B.Get_mpi() == mpi_graph_comm);
    
    Testing("OOMPI_Graph_comm::copy constructor");
    OOMPI_Graph_comm C(B);
    Test (C.Get_mpi() == mpi_graph_comm);
    
    Testing("OOMPI_Graph_comm::assignment");
    D = B;
    Test (D.Get_mpi() == mpi_graph_comm);
  }
  
  Pass();

  Testing("OOMPI_Graph_comm::Dup()");
  OOMPI_Graph_comm E;
  E = D.Dup();
  Test (E.Compare(D) == OOMPI_CONGRUENT);
  {
    int nnodes_mpi, nedges_mpi;
    int nnodes, nedges;

    Testing("OOMPI_Graph_comm::Num_nodes()");
    MPI_Graphdims_get(mpi_graph_comm, &nnodes_mpi, &nedges_mpi);
    Test(D.Num_nodes() == nnodes_mpi); 
    
    Testing("OOMPI_Graph_comm::Num_edges()");
    Test(D.Num_edges() == nedges_mpi); 

    Testing("OOMPI_Graph_comm::Dims_get");
    D.Dims_get(nnodes, nedges);
    Test (nnodes == nnodes_mpi && nedges == nedges_mpi);
  }
  
  {
    int nnodes, nedges, i; 
    D.Dims_get(nnodes, nedges);
    int *index = new int [nnodes], *edges = new int [nedges];
    int *pindex, *pedges;

    Testing("OOMPI_Graph_comm::Get()");
    if (flags[SKIP_SGI20]) 
      Done("Skipped (SGI 2.0)");
    else {
      D.Get(nnodes, nedges, index, edges);
      Midtest(nnodes == x_nnodes && nedges == x_nedges, "Graph_get 0");
      for (i=0; i< nnodes; i++) 
	Midtest(index[i] == x_index[i], "Graph_get 1a");
      for (i=0; i< nedges; i++)
	Midtest(edges[i] == x_edges[i], "Graph_get 1b");
      Sync();
    
      D.Get(index, edges);
      for (i=0; i< nnodes; i++) 
	Midtest(index[i] == x_index[i], "Graph_get 2a");
      for (i=0; i< nedges; i++)
	Midtest(edges[i] == x_edges[i], "Graph_get 2b");
      Pass();
    }

    Testing("OOMPI_Graph_comm::Get_index()");
    if (flags[SKIP_SGI20])
      Done("Skipped (SGI 2.0)");
    else {
      D.Get_index(index);
      pindex = D.Get_index();
      for (i=0; i<nnodes; i++)
	Midtest(index[i] == x_index[i], "Graph_get_index 1a");
      for (i=0; i<nnodes; i++)
	Midtest(pindex[i] == x_index[i], "Graph_get_index 1b");
      Pass();

      delete[] pindex;
    }
    
    Testing("OOMPI_Graph_comm::Get_edges()");
    if (flags[SKIP_SGI20])
      Done("Skipped (SGI 2.0)");
    else {
      D.Get_edges(edges);
      pedges = D.Get_edges();
      for (i=0; i<nedges; i++)
	Midtest(edges[i] == x_edges[i], "Graph_get_edges 1a");
      for (i=0; i<nedges; i++)
	Midtest(pedges[i] == x_edges[i], "Graph_get_edges 1b");
      Pass();

      delete[] pedges;
      delete[] x_index;
      delete[] x_edges;
      delete[] index;
      delete[] edges;
    }
  }
  
  Testing("OOMPI_Graph_comm::Neighbors()");
  {
    int Ns[100], *Nsp;
    D.Neighbors(1, D.Rank(), Ns);
    if (Ns[0] != (my_rank+1) % comm_size && 
	Ns[0] != (my_rank-1+comm_size) % comm_size)
      Fail();

    D.Neighbors(my_rank, Ns);
    if (Ns[0] !=  (my_rank+1) % comm_size && 
	Ns[0] != (my_rank-1+comm_size) % comm_size)
      Fail();

    D.Neighbors(Ns);
    if (Ns[0] != (my_rank+1) % comm_size && 
	Ns[0] != (my_rank-1+comm_size) % comm_size)
      Fail();

    Nsp = D.Neighbors();
    if (Nsp[0] != (my_rank+1) % comm_size && 
	Nsp[0] != (my_rank-1+comm_size) % comm_size)
      Fail();
    delete[] Nsp;

    Nsp = D.Neighbors(my_rank);
    if (Nsp[0] != (my_rank+1) % comm_size &&
	Nsp[0] != (my_rank-1+comm_size) % comm_size)
      Fail();
    delete[] Nsp;

    Nsp = D.Neighbors(1,D.Rank());
    if (Nsp[0] != (my_rank+1) % comm_size &&
	Nsp[0] != (my_rank-1+comm_size) % comm_size)
      Fail();
    delete[] Nsp;
  }
  Pass();
  
  Testing("OOMPI_Graph_comm::Neighbors_count()");
  {
    if (D.Neighbors_count(D.Rank()) != 2 ||
	D.Neighbors_count() != 2)
      Fail();
  }
  Pass();

  Testing("OOMPI_Graph_comm::Test_inter()");
  Test (D.Test_inter() == false);

  Testing("OOMPI_Intra_comm::Topo_test()");
  Test (D.Topo_test() == OOMPI_GRAPH);

  }

  MPI_Comm_free(&mpi_graph_comm);

  Pass();
}







