// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: messages.cc,v 1.6 2003/02/13 21:07:48 jsquyres Exp $
//
// OOMPI test program
// Test *all* OOMPI functionality
// 

#include <iostream>

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "oompi.h"
#include "oompi_test.h"


//
// Global vars for below
//

static int indent_level = 0;
static int column = 0;
static int num_bullets = -1;
static int waiting = 0;

static const char bullets[] = { '*', '-', 'o', '.', 0 };
static const int dest_column = 50;


//
// Local functions
//

static void check_for_failures(int my_code, const char* msg = 0);
static void Endline(const char* msg);
static inline void decrement() 
{ indent_level = (indent_level <= 0) ? 0 : indent_level - 1; };
static inline void increment()
{ indent_level++; };


//
// Testing
//
void
Testing(const char* message, MPI_Comm comm)
{
  int i;

  if (comm != MPI_COMM_NULL)
    MPI_Barrier(comm);

  if (my_rank != 0)
    return;

  if (num_bullets == -1)
    for (num_bullets= 0; bullets[num_bullets] != 0; num_bullets++)
      continue;

  // Check to see if we are waiting at the end of a line already

  if (waiting)
    std::cout << std::endl;

  // Output the indenting and bullet

  column= 0;
  for (i= 0; i < indent_level; i++) {
    std::cout <<"  ";
    column += 2;
  }

  // Output the message 

  Push(message);
  if (indent_level == 0)
    std::cout << std::endl;
  std::cout << bullets[indent_level % num_bullets] << " " << message << "... ";

  column += 6 + strlen(message);
  for (i= column ; i < dest_column; i++)
    std::cout << " ";
  std::cout.flush();

  waiting= 1;
  increment();
}


//
// Pass
//
void
Pass(const char* msg)
{
  check_for_failures(0);

  decrement();
  Endline(msg);
}


//
// Sync
//
void
Sync(const char* msg)
{
  check_for_failures(0, msg);
}


//
// Postpone
//
void
Postpone(const char* class_name)
{
  static char buffer[1024];

  decrement();
  sprintf(buffer, "POSTPONED -- %s", class_name);
  Endline(buffer);
  std::cout.flush();
}


//
// Done
//
void
Done(const char* msg)
{
  decrement();
  Endline(msg);
  std::cout.flush();
}


//
// Fail
//
void
Fail(const char* msg)
{
  check_for_failures(1, msg);
}


//
// Abort
//
void
Abort(const char* msg)
{
  Endline("FAIL");

  std::cerr << std::endl;
  if (msg != 0)
    std::cerr << "OOMPI test suite: " << msg << std::endl;

  std::cerr << "OOMPI test suite: major error!" << std::endl;
  std::cerr << "OOMPI test suite: attempting to abort..." << std::endl;

  MPI_Abort(MPI_COMM_WORLD, -1);

  // Shouldn't reach here

  if (my_rank <= 0)
    std::cerr << "OOMPI test suite: terminated" << std::endl << std::endl;

  exit(OOMPI_errno);
}


//
// Endline
//
static void
Endline(const char* msg)
{
  if (my_rank != 0)
    return;

  std::cout.flush();
  if (waiting) {
    std::cout << msg << std::endl;
    Pop();
  } else {
    // Output the indenting and bullet
    
    int i;
    column= 0;
    for (i= 0; i < indent_level; i++) {
      std::cout <<"  ";
      column += 2;
    }
    
    char *line= Pop();
    int len= (line != 0) ? strlen(line) : 0;
    std::cout << bullets[indent_level % num_bullets] << " " << line << "... ";
    for (i= column + 6 + len; i < dest_column; i++)
      std::cout << " ";
    std::cout << msg << std::endl;
    std::cout.flush();
  }

  waiting= 0;
  column= 0;
}


//
// Check for failures among ranks
//
void
check_for_failures(int my_code, const char* msg)
{
  static int num_fails;
  static int recv;
  static MPI_Status status;
  static int i;

  MPI_Allreduce(&my_code, &num_fails, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  // Did someone fail? 

  if (num_fails == 0)
    return;

  // Yes, someone failed 

  if (my_rank > 0 && my_code == 1) 
    MPI_Send(&my_rank, 1, MPI_INT, 0, 1234, MPI_COMM_WORLD);

  if (my_rank == 0) {
    // End the line with a FAIL, because someone failed
    
    Endline("FAIL");
    
    // If we have a descriptive message, print it
    
    std::cerr << std::endl;
    if (msg != 0)
      std::cerr << "OOMPI test suite: " << msg << std::endl;
    
    // Print right header message
    
    if (num_fails == 1)
      std::cerr << "OOMPI test suite: attempting to determine which rank failed..." << std::endl;
    else if (num_fails < comm_size)
      std::cerr << "OOMPI test suite: attempting to determine which " << num_fails << " ranks failed..." << std::endl;
    else
      std::cerr << "OOMPI test suite: all ranks failed" << std::endl;
    
    // Was I was of the failures (this could only be rank 0)?
    
    if (my_code == 1 && num_fails < comm_size)
      std::cerr << "OOMPI test suite: rank 0 failed" << std::endl;
    
    // Receive all the failure messages, print if num_fails < comm_size
    
    for (i = 0; i < num_fails - my_code; i++) {
      MPI_Recv(&recv, 1, MPI_INT, MPI_ANY_SOURCE, 1234, MPI_COMM_WORLD, 
	       &status);
      if (num_fails < comm_size)
	std::cerr << "OOMPI test suite: rank " << recv << " failed" << std::endl;
    }
    
    std::cerr << "OOMPI test suite: minor error" << std::endl;
    std::cerr << "OOMPI test suite: attempting to finalize..." << std::endl;
  }
  
  // We can only hope that this works...
  OOMPI_COMM_WORLD.Finalize();
  
  if (my_rank == 0)
    std::cerr << "OOMPI test suite: terminated" << std::endl << std::endl;
  
  exit(OOMPI_errno);
}
