// -*- c++ -*-
//
// Copyright (c) 2002-2003 Indiana University.  All rights reserved.
// Copyright (c) 1996, 1997, 1998, 2000 University of Notre Dame.
//                         All rights reserved.
// 
// This file is part of the OOMPI software package.  For license
// information, see the LICENSE file in the top level directory of the
// OOMPI source distribution.
//
// $Id: port_test.cc,v 1.17 2003/02/13 21:07:48 jsquyres Exp $
//
// OOMPI test program
// Test OOMPI_Port functionality
//

#include <iostream>

#include "oompi.h"
#include "oompi_test.h"


//
// Convenience function
//
#include <stdio.h>
static void inline
Testarray(double* a, double* b, int size, const char* msg = 0)
{
  int i;

  for (i = 0; i < size; i++)
    if (!doublecmp(a[i], b[i])) {
      std::cout << "Node: " << my_rank << " failed in testarray, a[" << i 
		<< "]=" << a[i] << " != b[" << i << "]=" << b[i] << std::endl;
      if (msg != 0)
	Fail(msg);
    }
  Sync();
};


//
// User object
//
class double_object : public OOMPI_User_type {
public:
  inline double_object() 
    : OOMPI_User_type(type, this, 201)
  { if (!type.Built()) type.Contiguous(data, 100); };

  double data[100];

private:
  static OOMPI_Datatype type;
};


//
// Instances of types
//
OOMPI_Datatype double_object::type;

void OOMPI_Port_test_constructors();
void OOMPI_Port_test_indexing();
void OOMPI_Port_test_pt2pt();
void OOMPI_Port_test_collective();
void OOMPI_Port_test_rest();

//
// OOMPI_Port_test
//
void
OOMPI_Port_test()
{
  Testing("OOMPI_Port");

  int i;
  char *yow(0);
  int size = 10240;

  OOMPI_COMM_WORLD.Set_error_action(OOMPI_ERRORS_ARE_FATAL);

  // Buffer that is needed for bsend(), etc.

  if (!flags[SKIP_IBM21014] && !flags[SKIP_IBM21015] && 
      !flags[SKIP_IBM21016] && !flags[SKIP_IBM21017]) {
    size = 10240;
    yow = new char[size];
    MPI_Buffer_attach(yow, size);
  }

  // Rank

  Testing("OOMPI_Port::Rank()");
  Test(OOMPI_COMM_WORLD.Rank() == my_rank);

  OOMPI_Port_test_constructors();
  OOMPI_Port_test_indexing();
  OOMPI_Port_test_pt2pt();
  OOMPI_Port_test_collective();
  OOMPI_Port_test_rest();

  if (!flags[SKIP_IBM21014] && !flags[SKIP_IBM21015] &&
      !flags[SKIP_IBM21016] && !flags[SKIP_IBM21017]) {
    MPI_Buffer_detach(yow, &i);
    delete[] yow;
  }

  Pass();
}

void OOMPI_Port_test_constructors() {
  // Default constructor

  Testing("OOMPI_Port::Constructors");
  {
    Testing("OOMPI_Port::Default constructor");
    OOMPI_Port A;
    int rank = A.Rank();
    Test(rank == OOMPI_PROC_NULL);
    
    // Copy constructor
    
    Testing("OOMPI_Port::Copy constructor");
    OOMPI_Port B(A);
    Test(B.Rank() == A.Rank());
    
    // MPI constructor
    
    Testing("OOMPI_Port::MPI constructor");
    OOMPI_Port C(MPI_COMM_WORLD, my_rank);
    Test(C.Rank() == my_rank);
    
    // operator ==
    
    Testing("OOMPI_Port::operator=()");
    OOMPI_Port D = C;
    Test(D.Rank() == my_rank);
  }
  Pass(); // Constructors

  Testing("OOMPI_Port::Destructor");
  {
  }
  Pass();
}

void OOMPI_Port_test_indexing() {

  // OOMPI_Comm class
  // OOMPI_Comm::operator[]

  Testing("OOMPI_Intra_comm::OOMPI_Comm class");
  {
    Testing("OOMPI_Comm::operator[]()");
    OOMPI_Port E = OOMPI_COMM_WORLD[my_rank];
    Test(E.Rank() == my_rank);
  }
  Pass(); // OOMPI_Comm
}

void OOMPI_Port_test_pt2pt() {

  OOMPI_Port pto = OOMPI_COMM_WORLD[to];
  OOMPI_Port pfrom = OOMPI_COMM_WORLD[from];
  MPI_Status mpi_status;
  int i, j, l;
  const int tag = 201;
  
  // Point to point 
  // [potentially] Blocking calls

  Testing("OOMPI_Port::Point to point");
  {
    Testing("OOMPI_Port::Blocking");
    {
      int recv1, send1= 555;
      double recv2, send2= 7.77;
      int send_array[] = { 333, 444, 555 };
      int recv_array[3], recv_array2[3];
      
      // Standard mode
      
      Testing("OOMPI_Port::operator<<()/operator>>()");
      
      recv1= 0;
      recv2= 0.0;
      if (my_rank % 2 == 0) {
	pto << send1 << send2;
	pfrom >> recv1 >> recv2;
      }
      else {
	pfrom >> recv1 >> recv2;
	pto << send1 << send2;
      }
      Test(send1 == recv1 && send2 == recv2);
      
      Testing("OOMPI_Port::Send()/Recv()");
      recv1= 0;
      recv2= 0.0;
      recv_array[0]= recv_array[1]= recv_array[2]= 0;
      recv_array2[0]= recv_array2[1]= recv_array2[2]= 0;
      if (my_rank % 2 == 0) {
	pto.Send(send1);
	pto.Send(send2, 1234);
	pto.Send(send_array, 3);
	pto.Send(send_array, 3, 1234);
	
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
      }
      else {
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
	
	pto.Send(send1);
	pto.Send(send2, 1234);
	pto.Send(send_array, 3);
	pto.Send(send_array, 3, 1234);
      }
      Test(send1 == recv1 && send2 == recv2 && 
	   send_array[0] == recv_array[0] &&
	   send_array[1] == recv_array[1] &&
	   send_array[2] == recv_array[2] &&
	   send_array[0] == recv_array2[0] &&
	   send_array[1] == recv_array2[1] &&
	   send_array[2] == recv_array2[2]);
      
      // Synchronous mode
      
      Testing("OOMPI_Port::Ssend()");
      recv1= 0;
      recv2= 0.0;
      recv_array[0]= recv_array[1]= recv_array[2]= 0;
      recv_array2[0]= recv_array2[1]= recv_array2[2]= 0;
      if (my_rank % 2 == 0) {
	pto.Ssend(send1);
	pto.Ssend(send2, 1234);
	pto.Ssend(send_array, 3);
	pto.Ssend(send_array, 3, 1234);
	
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
      }
      else {
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
	
	pto.Ssend(send1);
	pto.Ssend(send2, 1234);
	pto.Ssend(send_array, 3);
	pto.Ssend(send_array, 3, 1234);
      }
      Test(send1 == recv1 && send2 == recv2 && 
	   send_array[0] == recv_array[0] &&
	   send_array[1] == recv_array[1] &&
	   send_array[2] == recv_array[2] &&
	   send_array[0] == recv_array2[0] &&
	   send_array[1] == recv_array2[1] &&
	   send_array[2] == recv_array2[2]);
      
      // Buffered mode

      Testing("OOMPI_Port::Bsend()");
      if (flags[SKIP_IBM21014])
	Done("Skipped (IBM 2.1.0.14)");
      else if (flags[SKIP_IBM21015])
	Done("Skipped (IBM 2.1.0.15)");
      else if (flags[SKIP_IBM21016])
	Done("Skipped (IBM 2.1.0.16)");
      else if (flags[SKIP_IBM21017])
	Done("Skipped (IBM 2.1.0.17)");
      else {
	recv1= 0;
	recv2= 0.0;
	recv_array[0]= recv_array[1]= recv_array[2]= 0;
	recv_array2[0]= recv_array2[1]= recv_array2[2]= 0;
	
	if (my_rank % 2 == 0) {
	  pto.Bsend(send1);
	  pto.Bsend(send2, 1234);
	  pto.Bsend(send_array, 3);
	  pto.Bsend(send_array, 3, 1234);
	  
	  pfrom.Recv(recv1);
	  pfrom.Recv(recv2, 1234);
	  pfrom.Recv(recv_array, 3);
	  pfrom.Recv(recv_array2, 3, 1234);
	}
	else {
	  pfrom.Recv(recv1);
	  pfrom.Recv(recv2, 1234);
	  pfrom.Recv(recv_array, 3);
	  pfrom.Recv(recv_array2, 3, 1234);
	  
	  pto.Bsend(send1);
	  pto.Bsend(send2, 1234);
	  pto.Bsend(send_array, 3);
	  pto.Bsend(send_array, 3, 1234);
	}
	Test(send1 == recv1 && send2 == recv2 && 
	     send_array[0] == recv_array[0] &&
	     send_array[1] == recv_array[1] &&
	     send_array[2] == recv_array[2] &&
	     send_array[0] == recv_array2[0] &&
	     send_array[1] == recv_array2[1] &&
	     send_array[2] == recv_array2[2]);
      }
      
      // Ready mode
      
      Testing("OOMPI_Port::Rsend()");
      recv1= 0;
      recv2= 0.0;
      recv_array[0]= 0;
      if (my_rank % 2 == 0) {
	do_work();
	pto.Rsend(send1);
	do_work();
	pto.Rsend(send2, 1234);
	do_work();
	pto.Rsend(send_array, 3);
	do_work();
	pto.Rsend(send_array, 3, 1234);
	
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
      }
      else {
	pfrom.Recv(recv1);
	pfrom.Recv(recv2, 1234);
	pfrom.Recv(recv_array, 3);
	pfrom.Recv(recv_array2, 3, 1234);
	
	do_work();
	pto.Rsend(send1);
	do_work();
	pto.Rsend(send2, 1234);
	do_work();
	pto.Rsend(send_array, 3);
	do_work();
	pto.Rsend(send_array, 3, 1234);
      }
      Test(send1 == recv1 && send2 == recv2 && 
	   send_array[0] == recv_array[0] &&
	   send_array[1] == recv_array[1] &&
	   send_array[2] == recv_array[2] &&
	   send_array[0] == recv_array2[0] &&
	   send_array[1] == recv_array2[1] &&
	   send_array[2] == recv_array2[2]);
      
      // Probe
      
      Testing("OOMPI_Port::Probe()");
      {
	if ((my_rank % 2) == 0) {
	  MPI_Send(send_array, 3, MPI_INT, to, 100, MPI_COMM_WORLD);
	  MPI_Send(send_array, 3, MPI_INT, to, 100, MPI_COMM_WORLD);
	}
	else {
	  OOMPI_Status stat;
	  
	  stat = pfrom.Probe(100);
	  MPI_Recv(recv_array, 3, MPI_INT,from,100, MPI_COMM_WORLD, 
		   &mpi_status);
	  stat = pfrom.Probe(OOMPI_ANY_TAG);
	  MPI_Recv(recv_array, 3, MPI_INT,from,100, MPI_COMM_WORLD, 
		   &mpi_status);
	}
	Pass();
      }
    }      
    Pass(); // Blocking mode
    
    // Non blocking
    // Isend() / Irecv()
    
    Testing("OOMPI_Port::Non-blocking");
    {
      MPI_Request request[2];
      MPI_Status status2[2];
      
      double dsend[100];
      double drecv[100];
      
      Testing("OOMPI_Port::Isend()");
      {
	OOMPI_Request R1;
	
	i = to;
	R1 = OOMPI_COMM_WORLD[to].Isend(my_rank, tag);
	MPI_Irecv(&i, 1, MPI_INT, from, tag, MPI_COMM_WORLD, &request[0]);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Midtest(i == from, "Scalar comparison failed");
      }
      
      {
	OOMPI_Request R1;
	
	for (i = 0; i < 100; i++) {
	  dsend[i] = i * 3.456;
	  drecv[i] = 0.0;
	}
	R1 = OOMPI_COMM_WORLD[to].Isend(dsend, 100, tag);
	MPI_Irecv(drecv, 100, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		  &request[0]);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Testarray(dsend, drecv, 100, "Array_message");
      }
      Pass();
      
      // Ibsend
      
      Testing("OOMPI_Port::Ibsend()");
      if (flags[SKIP_IBM21014])
	Done("Skipped (IBM 2.1.0.14)");
      else if (flags[SKIP_IBM21015])
	Done("Skipped (IBM 2.1.0.15)");
      else if (flags[SKIP_IBM21016])
	Done("Skipped (IBM 2.1.0.16)");
      else if (flags[SKIP_IBM21017])
	Done("Skipped (IBM 2.1.0.17)");
      else {
	{
	  OOMPI_Request R1;
	  
	  i = to;
	  R1 = OOMPI_COMM_WORLD[to].Ibsend(my_rank, tag);
	  MPI_Irecv(&i, 1, MPI_INT, from, tag, MPI_COMM_WORLD, &request[0]);
	  request[1] = R1.Get_mpi();
	  MPI_Waitall(2, request, status2);
	  Midtest(i == from, "Scalar comparison failed");
	}
	
	{
	  OOMPI_Request R1;
	  
	  for (i = 0; i < 100; i++)
	    drecv[i] = 0.0;
	  R1 = OOMPI_COMM_WORLD[to].Ibsend(dsend, 100, tag);
	  MPI_Irecv(drecv, 100, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		    &request[0]);
	  request[1] = R1.Get_mpi();
	  MPI_Waitall(2, request, status2);
	  Testarray(dsend, drecv, 100, "Array_message");
	}
	Pass();
      }
      
      // Irsend
      
      Testing("OOMPI_Port::Irsend()");
      {
	OOMPI_Request R1;
	
	i = to;
	MPI_Irecv(&i, 1, MPI_INT, from, tag, MPI_COMM_WORLD, &request[0]);
	do_work();
	R1 = OOMPI_COMM_WORLD[to].Irsend(my_rank, tag);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Midtest(i == from, "Scalar comparison failed");
      }
      
      {
	OOMPI_Request R1;
	
	for (i = 0; i < 100; i++)
	  drecv[i] = 0.0;
	MPI_Irecv(drecv, 100, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		  &request[0]);
	do_work();
	R1 = OOMPI_COMM_WORLD[to].Irsend(dsend, 100, tag);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Testarray(dsend, drecv, 100, "Array_message");
      }
      Pass();
      
      // Issend
      
      Testing("OOMPI_Port::Issend()");
      {
	OOMPI_Request R1;
	
	i = to;
	R1 = OOMPI_COMM_WORLD[to].Issend(my_rank, tag);
	MPI_Irecv(&i, 1, MPI_INT, from, tag, MPI_COMM_WORLD, &request[0]);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Midtest(i == from, "Scalar comparison failed");
      }
      
      {
	OOMPI_Request R1;
	
	for (i = 0; i < 100; i++)
	  drecv[i] = 0.0;
	R1 = OOMPI_COMM_WORLD[to].Issend(dsend, 100, tag);
	MPI_Irecv(drecv, 100, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		  &request[0]);
	request[1] = R1.Get_mpi();
	MPI_Waitall(2, request, status2);
	Testarray(dsend, drecv, 100, "Array_message");
      }
      Pass();

      // Irecv
      
      Testing("OOMPI_Port::Irecv()");
      {
	OOMPI_Request R1, R5;
	
	i = to;
	R1 = OOMPI_COMM_WORLD[to].Isend(my_rank, tag);
	R5 = OOMPI_COMM_WORLD[from].Irecv(i, tag);
	request[0] = R1.Get_mpi();
	request[1] = R5.Get_mpi();
	MPI_Waitall(2, request, status2);
	Midtest(i == from, "Scalar comparison failed");
      }
      
      {
	OOMPI_Request R1, R5;
	
	for (i = 0; i < 100; i++)
	  drecv[i] = 0.0;
	R1 = OOMPI_COMM_WORLD[to].Isend(dsend, 100, tag);
	R5 = OOMPI_COMM_WORLD[from].Irecv(drecv, 100, tag);
	request[0] = R1.Get_mpi();
	request[1] = R5.Get_mpi();
	MPI_Waitall(2, request, status2);
	Testarray(dsend, drecv, 100, "Array_message");
      }
      Pass();

      // Iprobe
      
      Testing("OOMPI_Port::Iprobe()");
      {
	int send_array[] = { 333, 444, 555 };
	int recv_array[3];
	
	if ((my_rank % 2) == 0) {
	  MPI_Send(send_array, 3, MPI_INT, to, 100, MPI_COMM_WORLD);
	  MPI_Send(send_array, 3, MPI_INT, to, 100, MPI_COMM_WORLD);
	}
	else {
	  OOMPI_Status stat;
	  int flag= 0;
	
	  while (flag == 0)
	    stat= pfrom.Iprobe(100, flag);
	  MPI_Recv(recv_array, 3, MPI_INT,from, 100, MPI_COMM_WORLD, 
		   &mpi_status);
	  
	  flag= 0;
	  while (flag == 0)
	    stat= pfrom.Iprobe(OOMPI_ANY_TAG, flag);
	  MPI_Recv(recv_array, 3, MPI_INT,from, 100, MPI_COMM_WORLD, 
		   &mpi_status);
	}
      }
      Pass();
    }
    Pass(); // Non-blocking
    
    // Persistent
    // Send_init
      
    Testing("OOMPI_Port::Persistent");
    {
      double dsend[200];
      double drecv[200];
      MPI_Request mpi_request[2];
      MPI_Status mpi_status2[2];
      
      Testing("OOMPI_Port::Send_init()");
      {
	OOMPI_Request r1;
	r1 = pto.Send_init(my_rank, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  j = 0;
	  MPI_Irecv(&j, 1, MPI_INT, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Midtest(j == from, "Message");
	}
	r1.Free();
      }
      
      {
	OOMPI_Request r1;
	for (i = 0; i < 200; i++)
	  dsend[i] = i * 5.678;
	r1 = pto.Send_init(dsend, 200, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  for (l = 0; l < 200; l++)
	    drecv[l] = 0.0;
	  MPI_Irecv(drecv, 200, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Testarray(dsend, drecv, 200, "Array_message");
	}
	r1.Free();
      }
      Pass();

      // Bsend_init
      
      Testing("OOMPI_Port::Bsend_init()");
      if (flags[SKIP_IBM21014])
	Done("Skipped (IBM 2.1.0.14)");
      else if (flags[SKIP_IBM21015])
	Done("Skipped (IBM 2.1.0.15)");
      else if (flags[SKIP_IBM21016])
	Done("Skipped (IBM 2.1.0.16)");
      else if (flags[SKIP_IBM21017])
	Done("Skipped (IBM 2.1.0.17)");
      else if (flags[SKIP_SGI32])
	Done("Skipped (SGI 3.2 64-bit)");
      else {
	{
	  OOMPI_Request r1;
	  r1 = pto.Bsend_init(my_rank, tag);
	  mpi_request[0] = r1.Get_mpi();
	  for (i = 0; i < 5; i++) {
	    j = 0;
	    MPI_Irecv(&j, 1, MPI_INT, from, tag, MPI_COMM_WORLD, 
		      &mpi_request[1]);
	    MPI_Start(&mpi_request[0]);
	    MPI_Waitall(2, mpi_request, mpi_status2);
	    
	    Midtest(j == from, "Message");
	  }
	  r1.Free();
	}

	{
	  OOMPI_Request r1;
	  for (i = 0; i < 200; i++)
	    dsend[i] = i;
	  r1 = pto.Bsend_init(dsend, 200, tag);
	  mpi_request[0] = r1.Get_mpi();
	  for (i = 0; i < 5; i++) {
	    for (l = 0; l < 200; l++)
	      drecv[l] = 0.0;
	    MPI_Irecv(drecv, 200, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		      &mpi_request[1]);
	    MPI_Start(&mpi_request[0]);
	    MPI_Waitall(2, mpi_request, mpi_status2);
	    
	    Testarray(dsend, drecv, 200, "Array_message");
	  }
	  r1.Free();
	  Pass();
	}
      }
    
      // Rsend_init
      
      Testing("OOMPI_Port::Rsend_init()");
      {
	OOMPI_Request r1;
	r1 = pto.Rsend_init(my_rank, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  j = 0;
	  MPI_Irecv(&j, 1, MPI_INT, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  do_work();
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Midtest(j == from, "Message");
	}
	r1.Free();
      }
      
      {
	OOMPI_Request r1;
	for (i = 0; i < 200; i++)
	  dsend[i] = i * 5.678;
	r1 = pto.Rsend_init(dsend, 200, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  for (l = 0; l < 200; l++)
	    drecv[l] = 0.0;
	  MPI_Irecv(drecv, 200, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  do_work();
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Testarray(dsend, drecv, 200, "Array_message");
	}
	r1.Free();
      }
      Pass();
      
      // Ssend_init
      
      Testing("OOMPI_Port::Ssend_init()");
      {
	OOMPI_Request r1;
	r1 = pto.Ssend_init(my_rank, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  j = 0;
	  MPI_Irecv(&j, 1, MPI_INT, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	    MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Midtest(j == from, "Message");
	}
	r1.Free();
      }
      
      {
	OOMPI_Request r1;
	for (i = 0; i < 200; i++)
	  dsend[i] = i * 5.678;
	r1 = pto.Ssend_init(dsend, 200, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  for (l = 0; l < 200; l++)
	    drecv[l] = 0.0;
	  MPI_Irecv(drecv, 200, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Testarray(dsend, drecv, 200, "Array_message");
	}
	r1.Free();
      }
      Pass();
      
      // Recv_init
      
      Testing("OOMPI_Port::Recv_init()");
      {
	OOMPI_Request r1;
	r1 = pfrom.Recv_init(j, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  j = 0;
	  MPI_Isend(&my_rank, 1, MPI_INT, to, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Midtest(j == from, "Message");
	}
	r1.Free();
      }
      
      {
	OOMPI_Request r1;
	for (i = 0; i < 200; i++)
	  dsend[i] = i * 5.678;
	r1 = pfrom.Recv_init(drecv, 200, tag);
	mpi_request[0] = r1.Get_mpi();
	for (i = 0; i < 5; i++) {
	  for (l = 0; l < 200; l++)
	    drecv[l] = 0.0;
	  MPI_Isend(dsend, 200, MPI_DOUBLE, to, tag, MPI_COMM_WORLD, 
		    &mpi_request[1]);
	  MPI_Start(&mpi_request[0]);
	  MPI_Waitall(2, mpi_request, mpi_status2);
	  
	  Testarray(dsend, drecv, 200, "Array_message");
	}
	r1.Free();
      }
      Pass();
    }
    Pass(); // Persistent
  }
  Pass(); // Point to point
}

void OOMPI_Port_test_collective() {
  
  int i, j, l;

  // Rooted collective communication
  // Broadcast
  
  Testing("OOMPI_Port::Rooted collective");  
  {
    double recv2, send2 = 1234.56789;
    int send_array[] = { 333, 444, 555 };
    int recv_array[3];
    OOMPI_Port root = OOMPI_COMM_WORLD[0];
    
    Testing("OOMPI_Port::Broadcast()");
    if (my_rank == 0) {
      recv2 = send2;
      for (i = 0; i < 3; i++)
	recv_array[i] = send_array[i];
    }
    else {
      recv2 = 0;
      recv_array[0] = recv_array[1] = recv_array[2] = 0;
    }
    
    root.Bcast(recv2);
    Midtest(send2 == recv2, "Scalar broadcast fail");
    
    root.Bcast(recv_array, 3);
    Test(recv_array[0] == send_array[0] &&
	 recv_array[1] == send_array[1] &&
	 recv_array[2] == send_array[2], "Array broadcast fail");
    
    // Scatter
    
    Testing("OOMPI_Port::Scatter()");
    {
      const int size2 = 2;
      int *scat_buf = new int[comm_size * size2];
      int *my_scat_buf = new int[size2];
      for (i = 0, l = 0; i < comm_size; i++) {
	scat_buf[l++] = (i + 1);
	scat_buf[l++] = (i + 1) * 2;
      }
      
      // Scatter: Test one
      
      OOMPI_Message scat_msg(scat_buf, size2);
      OOMPI_Message my_scat_msg(my_scat_buf, size2);
      root.Scatter(scat_msg, my_scat_msg);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Uniform Message / Message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test two
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      if (my_rank == 0)
	root.Scatter(scat_msg, my_scat_msg);
      else
	root.Scatter(my_scat_msg);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Nonuniform Message / Message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test three
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      root.Scatter(scat_buf, size2, my_scat_msg);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Uniform Array_message / Message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test four
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      if (my_rank == 0)
	root.Scatter(scat_buf, size2, my_scat_msg);
      else
	root.Scatter(my_scat_msg);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Nonuniform Array_message / Message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test five
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      root.Scatter(scat_msg, my_scat_buf, size2);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Uniform Message / Array_message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test six
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      if (my_rank == 0)
	root.Scatter(scat_msg, my_scat_buf, size2);
      else
	root.Scatter(my_scat_buf, size2);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Nonuniform Message / Array_message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test seven
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      root.Scatter(scat_buf, size2, my_scat_buf, size2);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Uniform Array_message / Array_message");
      MPI_Barrier(MPI_COMM_WORLD);
      
      // Scatter: Test eight
      
      my_scat_buf[0] = my_scat_buf[1] = -1;
      if (my_rank == 0)
	root.Scatter(scat_buf, size2, my_scat_buf, size2);
      else
	root.Scatter(my_scat_buf, size2);
      Midtest(my_scat_buf[0] == (my_rank + 1) && 
	      my_scat_buf[1] == (my_rank + 1) * 2, 
	      "Nonuniform Array_message / Array_message");
      
      delete[] scat_buf;
      delete[] my_scat_buf;
    }    
    Pass();

    // Scatterv
    
    Testing("OOMPI_Port::Scatterv()");
    {
      int *counts = new int[comm_size];
      int *displs = new int[comm_size];
      int sum = 0;
      for (i = 0; i < comm_size; i++)
	sum += i + 1;
      int *gatherv_buf = new int[sum];
      int *scatterv_buf = new int[sum];
      for (j = 0, i = 0; i < comm_size; i++)
	for (l = 0; l <= i; l++, j++)
	  gatherv_buf[j] = l;
      
      // Scatterv: Test one

      for (i = 0; i < comm_size; i++) {
	counts[i] = 1;
	displs[i] = i;
	scatterv_buf[i] = i;
      }
      i = -1;
      root.Scatterv(scatterv_buf, counts, displs, i);
      Midtest(i == my_rank, "Uniform scalar scatterv fail");
      
      // Scatterv: Test two
      
      i = -1;
      if (my_rank == 0)
	root.Scatterv(scatterv_buf, counts, displs, i);
      else
	root.Scatterv(i);
      Midtest(i == my_rank, "Nonuniform scalar scatterv fail");
      
      // Scatterv: Test three
      
      for (l = 0, i = 0; i < comm_size; i++) {
	counts[i] = i + 1;
	displs[i] = l;
	l += i + 1;
      }
      for (j = 0, i = 0; i < comm_size; i++)
	for (l = 0; l <= i; l++, j++)
	  scatterv_buf[j] = i;
      for (i = 0; i < sum; i++)
	gatherv_buf[i] = 0;
      root.Scatterv(scatterv_buf, counts, displs, gatherv_buf, my_rank + 1);
      for (i = 0; i <= my_rank; i++)
	if (gatherv_buf[i] != my_rank)
	  Fail("Uniform array scatterv fail");
      Sync();
      
      // Scatterv: Test four
      
      for (i = 0; i < sum; i++)
	gatherv_buf[i] = 0;
      if (my_rank == 0)
	root.Scatterv(scatterv_buf, counts, displs, gatherv_buf, my_rank + 1);
      else
	root.Scatterv(gatherv_buf, my_rank + 1);
      for (i = 0; i <= my_rank; i++)
	if (gatherv_buf[i] != my_rank)
	  Fail("Uniform array scatterv fail");
      
      delete[] scatterv_buf;
      delete[] gatherv_buf;
      delete[] counts;
      delete[] displs;
    }
    Pass();
    
    // Gather

    Testing("OOMPI_Port::Gather()");
    {
      const int size2 = 2;
      double *coll_buf = new double[comm_size * size2];
      double send_buf[size2];
      send_buf[0] = my_rank;
      send_buf[1] = (my_rank + 1.2453) * 9.4823;
      
      // Gather: Test one

      OOMPI_Message send_msg(send_buf, size2);
      OOMPI_Message coll_msg(coll_buf, size2);
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      root.Gather(send_msg, coll_msg);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Message / Message");
      else
	Sync();
      
      // Gather: Test two
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Gather(send_msg, coll_msg);
      else
	root.Gather(send_msg);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Message / Message");
      else
	Sync();
      
      // Gather: Test three
      
      for (i = 0; i < comm_size * size2; i++)
	coll_buf[i] = 0.0;
      root.Gather(send_msg, coll_buf, size2);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Message / Array_message");
      else
	Sync();
      
      // Gather: Test four
      
      for (i = 0; i < comm_size * size2; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Gather(send_msg, coll_buf, size2);
      else
	root.Gather(send_msg);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Non-uniform Message / Array_message");
      else
	Sync();
      
      // Gather: Test five
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      root.Gather(send_buf, size2, coll_msg);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Array_message / Message");
      else
	Sync();
      
      // Gather: Test six
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Gather(send_buf, size2, coll_msg);
      else
	root.Gather(send_msg);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Non-uniform Array_message / Message");
      else
	Sync();
      
      // Gather: Test seven
      
      for (i = 0; i < comm_size * size2; i++)
	coll_buf[i] = 0.0;
      root.Gather(send_buf, size2, coll_buf, size2);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Array_message / Array_message");
      else
	Sync();
      
      // Gather: Test eight
      
      for (i = 0; i < comm_size * size2; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Gather(send_buf, size2, coll_buf, size2);
      else
	root.Gather(send_buf, size2);
      if (my_rank == 0)
	Testarray(coll_buf, send_buf, size2, 
		  "Uniform Array_message / Array_message");
      else
	Sync();
      
      delete[] coll_buf;
    }
    Pass();
    
    // Gatherv
    
    Testing("OOMPI_Port::Gatherv()");
    {
      int *counts = new int[comm_size];
      int *displs = new int[comm_size];
      int sum = 0;
      for (i = 0; i < comm_size; i++)
	sum += i + 1;
      int *gatherv_buf = new int[sum];
      int *scatterv_buf = new int[sum];
      for (j = 0, i = 0; i < comm_size; i++)
	for (l = 0; l <= i; l++, j++)
	  gatherv_buf[j] = l;
      
      // Gatherv: Test one
      
      for (i = 0; i < comm_size; i++) {
	counts[i] = 1;
	displs[i] = i;
      }
      for (i = 0; i < sum; i++)
	gatherv_buf[i] = 0;
      root.Gatherv(my_rank, gatherv_buf, counts, displs);
      if (my_rank == 0)
	for (i = 0; i < comm_size; i++)
	  if (gatherv_buf[i] != i)
	    Fail("Uniform scalar gatherv fail");
      Sync();
      
      // Gatherv: Test two
      
      for (i = 0; i < sum; i++)
	gatherv_buf[i] = 0;
      if (my_rank == 0)
	root.Gatherv(my_rank, gatherv_buf, counts, displs);
      else
	root.Gatherv(my_rank);
      if (my_rank == 0)
	for (i = 0; i < comm_size; i++)
	  if (gatherv_buf[i] != i)
	    Fail("Nonuniform scalar gatherv fail");
      Sync();
      
      // Gatherv: Test 3
      
      for (l = 0, i = 0; i < comm_size; i++) {
	counts[i] = i + 1;
	displs[i] = l;
	l += i + 1;
      }
      for (i = 0; i <= my_rank; i++)
	scatterv_buf[i] = my_rank;
      root.Gatherv(scatterv_buf, my_rank + 1, gatherv_buf, counts, displs);
      if (my_rank == 0)
	for (j = 0, i = 0; i < comm_size; i++)
	  for (l = 0; l <= i; l++, j++)
	    if (gatherv_buf[j] != i)
	      Fail("Uniform array gatherv fail");
      Sync();
      
      // Gatherv: Test 4
      
      for (i = 0; i < sum; i++)
	gatherv_buf[i] = 0;
      if (my_rank == 0)
	root.Gatherv(scatterv_buf, my_rank + 1, gatherv_buf, counts, displs);
      else
	root.Gatherv(scatterv_buf, my_rank + 1);
      if (my_rank == 0)
	for (j = 0, i = 0; i < comm_size; i++)
	  for (l = 0; l <= i; l++, j++)
	    if (gatherv_buf[j] != i)
	      Fail("Nonuniform array gatherv fail");
      Sync();
      
      delete[] gatherv_buf;
      delete[] scatterv_buf;
      delete[] counts;
      delete[] displs;
    }
    Pass();
    
    // Reduce
    
    Testing("OOMPI_Port::Reduce()");
    {
      double *coll_buf = new double [comm_size * comm_size];
      double *send_buf = new double [comm_size];
      double *sum = new double [comm_size];
      for (i = 0; i < comm_size; i++) {
	send_buf[i] = (i + 1);
	sum[i] = send_buf[i] * comm_size;
      }
      
      // Reduce: Test one
      
      OOMPI_Message send_msg(send_buf, comm_size);
      OOMPI_Message coll_msg(coll_buf, comm_size);
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;

      root.Reduce(send_msg, coll_msg, OOMPI_SUM);

      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Uniform Message / Message");
      else
	Sync();
      

      // Reduce: Test two

      for (i = 0; i < comm_size * comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Reduce(send_msg, coll_msg, OOMPI_SUM);
      else
	root.Reduce(send_msg, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Non-uniform Message / Message");
      else
	Sync();
      
      // Reduce: Test three
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      root.Reduce(send_msg, coll_buf, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Uniform Message / Array_message");
      else
	Sync();
      
      // Reduce: Test four

      for (i = 0; i < comm_size * comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Reduce(send_msg, coll_buf, OOMPI_SUM);
      else
	root.Reduce(send_msg, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Non-uniform Message / Array_message");
      else
	Sync();
      
      // Reduce: Test five
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      root.Reduce(send_buf, coll_msg, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Uniform Array_message / Message");
      else
	Sync();
      
      // Reduce: Test six

      for (i = 0; i < comm_size * comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Reduce(send_buf, coll_msg, OOMPI_SUM);
      else
	root.Reduce(send_buf, comm_size, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Non-uniform Array_message / Message");
      else
	Sync();
      
      // Reduce: Test seven
      
      for (i = 0; i < comm_size; i++)
	coll_buf[i] = 0.0;
      root.Reduce(send_buf, coll_buf, comm_size, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Uniform Array_message / Array_message");
      else
	Sync();
      
      // Reduce: Test eight

      for (i = 0; i < comm_size * comm_size; i++)
	coll_buf[i] = 0.0;
      if (my_rank == 0)
	root.Reduce(send_buf, coll_buf, comm_size, OOMPI_SUM);
      else
	root.Reduce(send_buf, comm_size, OOMPI_SUM);
      if (my_rank == 0)
	Testarray(coll_buf, sum, comm_size,
		  "Non-uniform Array_message / Array_message");
      else
	Sync();
      
      delete [] send_buf;
      delete [] sum;
      delete[] coll_buf;
    }
    Pass(); // Reduce
  }
  Pass(); // Rooted collective
}

void OOMPI_Port_test_rest() {

  int pos, i, j, k, l;
  const int tag = 201;

  // Intercomm create
  
  Testing("OOMPI_Port::Intercomm_create()");  
  Postpone("OOMPI_Inter_comm");
  
  // Datatypes
  // Contiguous

  Testing("OOMPI_Port::Datatypes");
  {
    OOMPI_Status stat;

    Testing("OOMPI_Datatype::Contiguous()");
    {
      double dsend[400];
      double drecv[400];
      OOMPI_Datatype contig1;
      contig1.Contiguous(dsend, 25);
      OOMPI_Message double_send_msg1(contig1, dsend, 4);
      OOMPI_Message double_recv_msg1(contig1, drecv, 4);
      
      for (i = 0; i < 100; i++) {
	dsend[i] = i * 1.2;
	drecv[i] = 0.0;
      }
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg1, to, tag, 
				       double_recv_msg1, from, tag);
      Testarray(dsend, drecv, 100, "Array_message");
      
      OOMPI_Datatype contig2;
      contig2.Contiguous(dsend[0], 25);
      OOMPI_Message double_send_msg2(contig2, dsend, 4);
      OOMPI_Message double_recv_msg2(contig2, drecv, 4);
      
      for (i = 0; i < 100; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg2, to, tag, 
				       double_recv_msg2, from, tag);
      Testarray(dsend, drecv, 100, "Message");
      
      OOMPI_Datatype contig3;
      contig3.Contiguous_type(OOMPI_DOUBLE, 25);
      OOMPI_Message double_send_msg3(contig3, dsend, 4);
      OOMPI_Message double_recv_msg3(contig3, drecv, 4);
      
      for (i = 0; i < 100; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg3, to, tag, 
				       double_recv_msg3, from, tag);
      Testarray(dsend, drecv, 100, "OOMPI_Datatype");
    }
    Pass();
    
    // Vector

    Testing("OOMPI_Datatype::Vector()");
    {
      double dsend[200];
      double drecv[200];
      
      // Fill the array in the *strange* pattern necessary for the
      // type_vector (i.e. no holes at the end of each vector)

      for (pos = i = 0; i < 5; i++)
	for (l = 0; l < 10; l++)
	  for (j = 0; j < 4; j++, pos++)
	    if (j < 2)
	      dsend[pos] = pos;
	    else if (l == 9)
	      pos--;
	    else
	      dsend[pos] = 0;

      // Test 1

      OOMPI_Datatype vector1;
      vector1.Vector(2, 4, dsend, 10);
      OOMPI_Message double_send_msg1(vector1, dsend, 5);
      OOMPI_Message double_recv_msg1(vector1, drecv, 5);

      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg1, to, tag, 
				       double_recv_msg1, from, tag);
      // NOTE-we really should check all the points in the vector
      Testarray(dsend, drecv, 100, "Array_message");

      // Test 2

      OOMPI_Datatype vector2;
      vector2.Vector(2, 4, dsend[0], 10);
      OOMPI_Message double_send_msg2(vector2, dsend, 5);
      OOMPI_Message double_recv_msg2(vector2, drecv, 5);
      
      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg2, to, tag, 
				       double_recv_msg2, from, tag);
      Testarray(dsend, drecv, 100, "Message");
      
      // Test 3

      OOMPI_Datatype vector3;
      vector3.Vector_type(2, 4, OOMPI_DOUBLE, 10);
      OOMPI_Message double_send_msg3(vector3, dsend, 5);
      OOMPI_Message double_recv_msg3(vector3, drecv, 5);
      
      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg3, to, tag, 
				       double_recv_msg3, from, tag);
      Testarray(dsend, drecv, 100, "OOMPI_Datatype");
    }
    Pass();
    
    // Hvector

    Testing("OOMPI_Datatype::Hvector()");
    {
      double dsend[200];
      double drecv[200];
      
      // Fill the array in the *strange* pattern necessary for the
      // type_hvector (i.e. no holes at the end of each hvector)

      for (pos = i = 0; i < 5; i++)
	for (l = 0; l < 10; l++)
	  for (j = 0; j < 4; j++, pos++)
	    if (j < 2)
	      dsend[pos] = pos * 2.345;
	    else if (l == 9)
	      pos--;
	    else
	      dsend[pos] = 0.0;

      // Test 1

      OOMPI_Datatype hvector1;
      hvector1.Hvector(2, 4 * sizeof(double), dsend, 10);
      OOMPI_Message double_send_msg1(hvector1, dsend, 5);
      OOMPI_Message double_recv_msg1(hvector1, drecv, 5);

      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg1, to, tag, 
				       double_recv_msg1, from, tag);
      Testarray(dsend, drecv, 100, "Message");

      // Test 2

      OOMPI_Datatype hvector2;
      hvector2.Hvector(2, 4 * sizeof(double), dsend[0], 10);
      OOMPI_Message double_send_msg2(hvector2, dsend, 5);
      OOMPI_Message double_recv_msg2(hvector2, drecv, 5);
      
      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg2, to, tag, 
				       double_recv_msg2, from, tag);
      Testarray(dsend, drecv, 100, "Array_message");
      
      // Test 3

      OOMPI_Datatype hvector3;
      hvector3.Hvector_type(2, 4 * sizeof(double), OOMPI_DOUBLE, 10);
      OOMPI_Message double_send_msg3(hvector3, dsend, 5);
      OOMPI_Message double_recv_msg3(hvector3, drecv, 5);
      
      for (i = 0; i < 200; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg3, to, tag, 
				       double_recv_msg3, from, tag);
      Testarray(dsend, drecv, 100, "OOMPI_Datatype");
    }
    Pass();
    
    // Indexed
    
    Testing("OOMPI_Datatype::Indexed()");
    if (flags[SKIP_CRAY1104])
      Done("Skipped (Cray 1.1.0.4)");
    else {
      OOMPI_Request req;

      // Create a datatype for the following pattern:
      // An array of doubles
      // There will be comm_size blocks, for values of n=0..(comm_size - 1)
      // Block n will contain (n + 1) doubles and one hole (the size of a
      //   double)
      
      const int size2 = 10;
      const int count = 5;
      int lengths[size2];
      int disps[size2];
      int sum;

      for (sum = i = 0; i < size2; i++) {
	lengths[i] = (i + 1);
	disps[i] = sum;
	sum += lengths[i] + 1;
      }

      double *dsend = new double[sum * count];
      double *drecv = new double[sum * count];
      for (pos = i = 0; i < count; i++)
	for (k = 0; k < size2; k++)
	  for (l = 0; l < (k + 2); l++, pos++)
	    if (l < (k + 1))
	      dsend[pos] = (pos + 1) * 6.789;
	    else {
	      dsend[pos] = 0.0;
	      if (k == size2 - 1)
		pos--;
	    }

      // Test 1

      OOMPI_Datatype indexed1;
      indexed1.Indexed(lengths, disps, dsend[0], size2);
      OOMPI_Message double_send_msg1(indexed1, dsend, count);
      OOMPI_Message double_recv_msg1(indexed1, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg1, to, tag, 
				       double_recv_msg1, from, tag);
      Testarray(dsend, drecv, pos, "Message");

      // Test 2

      OOMPI_Datatype indexed2;
      indexed2.Indexed(lengths, disps, dsend, size2);
      OOMPI_Message double_send_msg2(indexed2, dsend, count);
      OOMPI_Message double_recv_msg2(indexed2, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg2, to, tag, 
				       double_recv_msg2, from, tag);
      Testarray(dsend, drecv, pos, "Array_message");

      // Test 3

      OOMPI_Datatype indexed3;
      indexed3.Indexed_type(lengths, disps, OOMPI_DOUBLE, size2);
      OOMPI_Message double_send_msg3(indexed3, dsend, count);
      OOMPI_Message double_recv_msg3(indexed3, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg3, to, tag, 
				       double_recv_msg3, from, tag);
      Testarray(dsend, drecv, pos, "OOMPI_Datatype");

      delete[] dsend;
      delete[] drecv;
    }
    Pass();
    
    // Hindexed
    
    Testing("OOMPI_Datatype::Hindexed()");
    {
      OOMPI_Request req;

      // Create a datatype for the following pattern:
      // An array of doubles
      // There will be comm_size blocks, for values of n=0..(comm_size - 1)
      // Block n will contain (n + 1) doubles and one hole (the size of a
      //   double)
      
      const int size2 = 3;
      const int count = 2;
      int lengths[size2];
      OOMPI_Aint disps[size2];
      int sum;

      for (sum = i = 0; i < size2; i++) {
	lengths[i] = (i + 1);
	disps[i] = sum * sizeof(double);
	sum += lengths[i] + 1;
      }

      double *dsend = new double[sum * count];
      double *drecv = new double[sum * count];
      for (pos = i = 0; i < count; i++)
	for (k = 0; k < size2; k++)
	  for (l = 0; l < (k + 2); l++, pos++)
	    if (l < (k + 1))
	      dsend[pos] = (pos + 1) * 6.789;
	    else {
	      dsend[pos] = 0.0;
	      if (k == size2 - 1)
		pos--;
	    }

      // Test 1

      OOMPI_Datatype hindexed1;
      hindexed1.Hindexed(lengths, disps, dsend[0], size2);
      OOMPI_Message double_send_msg1(hindexed1, dsend, count);
      OOMPI_Message double_recv_msg1(hindexed1, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg1, to, tag, 
				       double_recv_msg1, from, tag);
      Testarray(dsend, drecv, pos, "Message");

      // Test 2

      OOMPI_Datatype hindexed2;
      hindexed2.Hindexed(lengths, disps, dsend, size2);
      OOMPI_Message double_send_msg2(hindexed2, dsend, count);
      OOMPI_Message double_recv_msg2(hindexed2, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg2, to, tag, 
				       double_recv_msg2, from, tag);
      Testarray(dsend, drecv, pos, "Array_message");

      // Test 3

      OOMPI_Datatype hindexed3;
      hindexed3.Hindexed_type(lengths, disps, OOMPI_DOUBLE, size2);
      OOMPI_Message double_send_msg3(hindexed3, dsend, count);
      OOMPI_Message double_recv_msg3(hindexed3, drecv, count);

      for (i = 0; i < pos; i++)
	drecv[i] = 0.0;
      stat = OOMPI_COMM_WORLD.Sendrecv(double_send_msg3, to, tag, 
				       double_recv_msg3, from, tag);
      Testarray(dsend, drecv, pos, "OOMPI_Datatype");

      delete[] dsend;
      delete[] drecv;
    }
    Pass();
  }
  Pass(); // Datatypes

  Testing("OOMPI_PORT_NULL");
  {
    bool caught = false;
    OOMPI_errno = MPI_SUCCESS;
    OOMPI_COMM_WORLD.Set_error_action(OOMPI_ERRORS_EXCEPTION);
#if OOMPI_HAVE_EXCEPTIONS
    try {
      int foo = 1;

      OOMPI_PORT_NULL >> foo;
      OOMPI_PORT_NULL << foo;
    }
    catch (OOMPI_Error err) {
      caught = true;
    }
#else
    int i = 1;
    OOMPI_PORT_NULL >> i;
    if (OOMPI_errno != OOMPI_SUCCESS)
      caught = true;
    else {
      OOMPI_PORT_NULL << i;
      if (OOMPI_errno != OOMPI_SUCCESS)
	caught = true;
    }
#endif
    Test(!caught);
  }
}
